@@ -100,6 +100,26 @@ notify_notification_set_hint_variant (NotifyNotification *notification,
                                                            key,
                                                            (guchar) h_byte);
                 }
+        } else if (g_ascii_strcasecmp (type, "boolean") == 0) {
+                gboolean h_boolean = FALSE;
+
+                if (g_ascii_strcasecmp (value, "true") == 0) {
+                        h_boolean = TRUE;
+                } else if (g_ascii_isdigit (*value)) {
+                        h_boolean = !!g_ascii_strtoull (value, NULL, 10);
+                }
+
+                notify_notification_set_hint (notification, key,
+                                              g_variant_new_boolean (h_boolean));
+        } else if (g_ascii_strcasecmp (type, "variant") == 0) {
+                GVariant *variant = g_variant_parse (NULL, value, NULL, NULL, NULL);
+
+                if (variant != NULL) {
+                        notify_notification_set_hint (notification, key, variant);
+                } else {
+                        conv_error = TRUE;
+                }
+
         } else {
                 *error = g_error_new (G_OPTION_ERROR,
                                       G_OPTION_ERROR_BAD_VALUE,
@@ -171,6 +191,7 @@ main (int argc, char *argv[])
         static gint         notification_id = 0;
         static gboolean     do_version = FALSE;
         static gboolean     hint_error = FALSE, show_error = FALSE;
+        static gboolean     transient = FALSE;
         static gboolean     wait = FALSE;
         static glong        expire_timeout = NOTIFY_EXPIRES_DEFAULT;
         GOptionContext     *opt_ctx;
@@ -195,9 +216,12 @@ main (int argc, char *argv[])
                 {"category", 'c', 0, G_OPTION_ARG_FILENAME, &type,
                  N_("Specifies the notification category."),
                  N_("TYPE[,TYPE...]")},
+                {"transient", 'e', 0, G_OPTION_ARG_NONE, &transient,
+                 N_("Create a transient notification"),
+                 NULL},
                 {"hint", 'h', 0, G_OPTION_ARG_FILENAME_ARRAY, &hints,
                  N_
-                 ("Specifies basic extra data to pass. Valid types are int, double, string and byte."),
+                 ("Specifies basic extra data to pass. Valid types are boolean, int, double, string, byte and variant."),
                  N_("TYPE:NAME:VALUE")},
                 {"print-id", 'p', 0, G_OPTION_ARG_NONE, &print_id,
                  N_ ("Print the notification ID."), NULL},
@@ -225,7 +249,7 @@ main (int argc, char *argv[])
 
         setlocale (LC_ALL, "");
 
-#if !GLIB_CHECK_VERSION (2, 36, 0)
+#ifndef GLIB_VERSION_2_36
         g_type_init ();
 #endif
 
@@ -282,6 +306,11 @@ main (int argc, char *argv[])
         notify_notification_set_timeout (notify, expire_timeout);
         notify_notification_set_app_name (notify, app_name);
 
+        if (transient) {
+                notify_notification_set_hint (notify, "transient",
+                                              g_variant_new_boolean (TRUE));
+        }
+
         g_free (body);
 
         /* Set hints */
